/*
 * oggSlideshow creates a slideshow from a number of pictures
 *
 * Copyright (C) 2008-2009 Joern Seger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef __WIN32
#define __GNU_LIBRARY__
#include "../win32/getopt_win.h"
#endif

#include <iostream>
#include <map>
#include <vector>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <ctime>
//#include <cc++/slog.h>

#include "definition.h"
#include "theoraEncoder.h"
#include "fileRepository.h"
#include "streamMux.h"
#include "videoFormats.h"

const std::string validChars("0123456789,x");

void extractUint32(std::deque<uint32>& list, const std::string& _argument,
    char seperator)
{
  std::string argument(_argument);
  std::stringstream str;
  std::string substr;

  // if there is no argument given, the first frame will be created as a thumbnail
  if (argument.empty()) {
    list.push_back(0);
    return;
  }

  // delete all invalid data
  std::size_t pos;
  while ((pos = argument.find_first_not_of(validChars)) != std::string::npos){
    std::cerr << "erasing <"<<argument.at(pos)<<">\n";
    argument.erase(pos,1);
  }

  str << argument;

  uint32 value(0);
  while (!str.eof()) {
    std::stringstream part;
    getline(str, substr, seperator);
    part << substr;
    part >> value;
    list.push_back(value);
  }
}

void th_clean_ycbcr(th_ycbcr_buffer& theoraPictureBuffer)
{
#ifdef HAVE_BZERO
  bzero(theoraPictureBuffer, sizeof(th_img_plane)*3);
#else
  for(int i(0);i<sizeof(th_img_plane)*3;++i) *(((uint8*)theoraPictureBuffer)+i) = 0;
#endif


}

void th_free_ycbcr(th_ycbcr_buffer& theoraPictureBuffer)
{
  // kind of ugly

  delete[] theoraPictureBuffer[0].data;
  delete[] theoraPictureBuffer[1].data;
  delete[] theoraPictureBuffer[2].data;

  th_clean_ycbcr(theoraPictureBuffer);
}

void KenBurnsSequence(StreamMux& streamCreate, TheoraEncoder& theora_encoder,
    RGBPlane& origPlane, uint32 width, uint32 height, uint32 fullLength,
    uint32 blindLength, float framesPerSecond)
{
  RGBPlane blackPlane(width, height);

  OggPacket packet;
  th_ycbcr_buffer theoraPictureBuffer;
  th_clean_ycbcr(theoraPictureBuffer);

  uint32 planesize = width*height*4; // 3 Colors + Alpha channel
  memset(blackPlane.plane, 0x00, planesize);

  float maxZoomfactor;
  if ((origPlane.width*1.0)/(width*1.0)*(height*1.0) < (origPlane.height*1.0))
    maxZoomfactor = (origPlane.width*1.0)/(width*1.0);
  else
    maxZoomfactor = (origPlane.height*1.0)/(height*1.0);

  float zoomStart = maxZoomfactor*0.75+rand()*(maxZoomfactor*0.25)/(RAND_MAX*1.0);
  float zoomEnd   = maxZoomfactor*0.75+rand()*(maxZoomfactor*0.25)/(RAND_MAX*1.0);

#ifdef DEBUG
  std::cerr << "Zooming ("<<maxZoomfactor<<") from factor "<<zoomStart<<" to "<<zoomEnd<<std::endl;
#endif

  /* calculate a random walk */
  float availableXStart = origPlane.width - width*zoomStart;
  float availableYStart = origPlane.height - height*zoomStart;

  float availableXEnd   = origPlane.width - width*zoomEnd;
  float availableYEnd   = origPlane.height - height*zoomEnd;

  if ((availableXStart < 0) || (availableYStart < 0) || (availableXEnd < 0) || (availableYEnd < 0))
    throw "KenBurnsSequence: picture to small\n";

  float startpointX(0);
  float startpointY(0);

  float endpointX(0);
  float endpointY(0);

  float availLengthSqr(powf((availableXStart-availableXEnd),2.0)+powf((availableYStart-availableYEnd),2.0));
  float lengthSqr;

  do {
    startpointX = (((float)rand())*availableXStart)/(RAND_MAX*1.0);
    startpointY = (((float)rand())*availableYStart)/(RAND_MAX*1.0);

    endpointX = (((float)rand())*availableXEnd)/(RAND_MAX*1.0);
    endpointY = (((float)rand())*availableYEnd)/(RAND_MAX*1.0);

    // calculate walklength
    float lengthX = fabs(startpointX - endpointX);
    float lengthY = fabs(startpointY - endpointY);

    lengthSqr = powf(lengthX,2.0) + powf(lengthY,2.0);

  } while (lengthSqr < availLengthSqr/4.0);

#ifdef DEBUG
  std::cerr << "Original Picture size: < "<< origPlane.width <<" | " << origPlane.height<<" >\n";

  std::cerr << "Running from point < "<<startpointX<<" | "<<startpointY<<" > to point < "
  <<endpointX<<" | "<<endpointY<<" >\n";

  std::cerr << "original frame size from < " << width*zoomStart << " | " << height*zoomStart <<" > to < "
  << width*zoomEnd <<" | "<< height*zoomEnd <<" >\n";
#endif

  float stepX(0), stepY(0), stepZoom(0);

  fullLength = (uint32)(fullLength * framesPerSecond); // calculate the frames
  blindLength = (uint32)(blindLength * framesPerSecond);

  uint32 length;
  if (fullLength < 2*blindLength)
    length = 0;
  else
    length = fullLength - 2*blindLength;

  float stepXWidth((endpointX - startpointX)/(fullLength*1.0));
  float stepYWidth((endpointY - startpointY)/(fullLength*1.0));

  float stepZoomWidth((zoomEnd - zoomStart)/(fullLength*1.0));

  /* create the blinding: black to picture */
  for (uint32 i(0); i<blindLength; ++i) {

    RGBPlane actPlane;
    RGBPlane subPlane;

    try {
      // get the picture to be fade in
      subPlane = origPlane.subframe(width, height, startpointX+stepX, startpointY+stepY, 1.0/(zoomStart+stepZoom));

      // calculate the next fader picture
      actPlane = subPlane.crossfade(blackPlane,(i*1.0)/(blindLength*1.0));

      // let the fading go on
      stepX += stepXWidth;
      stepY += stepYWidth;
      stepZoom += stepZoomWidth;

      /* create a frame that is understood by the encoder */
      actPlane.exportYCrCb_theora(theoraPictureBuffer);

      theora_encoder << theoraPictureBuffer;
      theora_encoder >> packet;
      std::cerr << "\r  " << std::fixed << packet.getPacketNo()/framesPerSecond<<"               ";
      streamCreate << packet;

    } catch (const char* data) {
      std::cout << data << std::endl;
      exit(-1);
    }

    actPlane.cleanup();
    subPlane.cleanup();

  }

  /* create pictures without any fading */
  for (uint32 i(0); i<length; ++i) {

    RGBPlane actPlane;

    try {
      actPlane = origPlane.subframe(width, height, startpointX+stepX, startpointY+stepY, 1.0/(zoomStart+stepZoom));

      stepX += stepXWidth;
      stepY += stepYWidth;
      stepZoom += stepZoomWidth;

      actPlane.exportYCrCb_theora(theoraPictureBuffer);

      theora_encoder << theoraPictureBuffer;
      theora_encoder >> packet;
      std::cerr << "\r  " << std::fixed << packet.getPacketNo()/framesPerSecond<<"               ";
      streamCreate << packet;

    } catch (const char* data) {
      std::cout << data<<std::endl;
      exit(-1);
    }

    actPlane.cleanup();

  }

  for (uint32 i(0); i<blindLength; ++i) {

    RGBPlane actPlane;
    RGBPlane subPlane;

    try {

      subPlane = origPlane.subframe(width, height, startpointX+stepX, startpointY+stepY, 1.0/(zoomStart+stepZoom));

      // calculate the next fader picture
      actPlane = blackPlane.crossfade(subPlane,(i*1.0)/(blindLength*1.0));

      stepX += stepXWidth;
      stepY += stepYWidth;
      stepZoom += stepZoomWidth;

      actPlane.exportYCrCb_theora(theoraPictureBuffer);

      theora_encoder << theoraPictureBuffer;
      theora_encoder >> packet;
      std::cerr << "\r  " << std::fixed << packet.getPacketNo()/framesPerSecond<<"               ";
      streamCreate << packet;

    } catch (const char* data) {
      std::cout << data<<std::endl;
      exit(-1);
    }

    actPlane.cleanup();
    subPlane.cleanup();

  }

  blackPlane.cleanup();

  th_free_ycbcr(theoraPictureBuffer);

}

void crossfadeSequence(StreamMux& streamCreate, TheoraEncoder& theora_encoder,
    RGBPlane& origPlane, uint32 width, uint32 height, uint32 fullLength,
    uint32 blindLength, float framesPerSecond)
{
  static bool first(true);
  static RGBPlane previousPlane;
  RGBPlane actPlane;

  OggPacket packet;
  th_ycbcr_buffer theoraPictureBuffer;
  th_clean_ycbcr(theoraPictureBuffer);

  if ((origPlane.width != width) || (origPlane.height != height)) {
    actPlane = origPlane.reframe(width, height);
  } else {
    // copy the data
    actPlane = RGBPlane(width, height); //origPlane;

    // copy the plane
    uint32 size(origPlane.width*origPlane.height*4);
    memcpy(actPlane.plane, origPlane.plane, size);
  }

  if (first == false) {

    uint32 steps((uint32)(blindLength*framesPerSecond));

    for (uint32 i(0); i< steps; ++i) {
      /* create a frame that is understood by the encoder */
      RGBPlane plane = actPlane.crossfade(previousPlane, (i*1.0)/(steps*1.0));
      plane.exportYCrCb_theora(theoraPictureBuffer);

      theora_encoder << theoraPictureBuffer;
      theora_encoder >> packet;
      std::cerr << "\r  "<< std::fixed << packet.getPacketNo()/framesPerSecond
          <<"               ";
      streamCreate << packet;

      plane.cleanup();

    }
  }

  first = false;

  uint32 steps((uint32)((fullLength-blindLength)*framesPerSecond));

  /* create a frame that is understood by the encoder */
  actPlane.exportYCrCb_theora(theoraPictureBuffer);

  for (uint32 i(0); i< steps; ++i) {

    theora_encoder << theoraPictureBuffer;
    theora_encoder >> packet;
    std::cerr << "\r  "<< std::fixed << packet.getPacketNo()/framesPerSecond
        <<"               ";
    streamCreate << packet;

  }

  previousPlane.cleanup();
  previousPlane = actPlane;
  th_free_ycbcr(theoraPictureBuffer);

}

void frameShowSequence(StreamMux& streamCreate, TheoraEncoder& theora_encoder,
    RGBPlane& origPlane, uint32 width, uint32 height, uint32 fullLength,
    uint32 blindLength, float framesPerSecond)
{
  RGBPlane actPlane;

  OggPacket packet;
  th_ycbcr_buffer theoraPictureBuffer;

  //  YCbCrPlane ycbcrPicture;
  th_clean_ycbcr(theoraPictureBuffer);

  if ((origPlane.width != width) || (origPlane.height != height)) {
    actPlane = origPlane.reframe(width, height);
  } else {
    // copy the data
    actPlane = RGBPlane(width, height);

    // copy the plane
    uint32 size(origPlane.width*origPlane.height*4);
    memcpy(actPlane.plane, origPlane.plane, size);
  }

  uint32 steps((uint32)(fullLength*framesPerSecond));

  /* create a frame that is understood by the encoder */
  actPlane.exportYCrCb_theora(theoraPictureBuffer);

  for (uint32 i(0); i< steps; ++i) {

    theora_encoder << theoraPictureBuffer;
    theora_encoder >> packet;
    std::cerr << "\r  "<< std::fixed << packet.getPacketNo()/framesPerSecond
        <<"               ";
    streamCreate << packet;

  }

  actPlane.cleanup();
  th_free_ycbcr(theoraPictureBuffer);
}

void printHelpScreen(std::string& name)
{
  std::cerr << "usage: "<< name <<" [options] <picture1.bmp> <picture2.bmp> \n";
  std::cerr << "Options: \n"
            << " -s <width>x<height>: picture width/height of the output frame\n"
            << " -f <frames/s>      : frames per second\n"
            << " -o <output file>   : name of the output file\n"
            << " -l <length>        : number of frames per picture frequence\n"
            << " -d <datarate>      : datarate in bit/second\n"
            << " -r <resample>      : resizes the original pictures to video frame width/height and the additional resample factor\n"
            << " -e                 : reframe picture\n"
            << " -t <type>          : kb - Ken Burns\n"
            << "                      cf - cross fade\n"
            << "                      p  - plain\n";
}

int main(int argc, char* argv[])
{
  enum seqType {
    seqType_KenBurns,
    seqType_Crossfade,
    seqType_Plain
  };

  uint32 width(480);
  uint32 height(320);
  uint32 framesPerSecond(24);
  std::string outputFile("diashow.ogg");
  uint32 length(8);
  uint32 datarate(256000);
  seqType showType(seqType_KenBurns);
  float resample(1.2);
  bool reframe(false);

  srand(time(0));

  std::string programName(argv[0]);

  int opt;
  while ((opt = getopt(argc, argv, "hp:f:o:l:d:r:t:s:e")) != EOF)

    switch (opt) {

    case 'h':
    case '?':
      printHelpScreen(programName);
      exit(-1);

    case 's': {
      std::deque<uint32> framesize;
      extractUint32(framesize, optarg, 'x');
      if (framesize.size() != 2) {
        std::cerr << "please specify the size in the following way: -s320x480\n";
        exit(-1);
      }
      width = framesize[0];
      height = framesize[1];

    } break;

    case 'f':
      framesPerSecond = atoi(optarg);
      break;

    case 'o':
      outputFile = std::string(optarg);
      break;

    case 'l':
      length = atoi(optarg); // yes, I know the atoi bug
      break;

    case 'd':
      datarate = atoi(optarg); // yes, I know the atoi bug
      break;

    case 'r':
      resample = atof(optarg);
      if ((resample < 1) || (resample > 2))
        resample = 1.2;
      break;

    case 'e': {
    	std::cerr << "reframing\n";
    	reframe = true;
        break;
    }

    case 't': {
      std::string typeStr(optarg);
      if ((typeStr == "kb") || (typeStr =="KenBurns")|| (typeStr == "KB")) {
        showType = seqType_KenBurns;
        break;
      }
      if ((typeStr == "cf") || (typeStr =="crossfade")) {
        showType = seqType_Crossfade;
        break;
      }

      if ((typeStr == "p") || (typeStr =="plain")|| (typeStr == "simple")) {
        showType = seqType_Plain;
        break;
      }

      std::cerr << "Unknown Type: using Ken Burns";
      showType = seqType_KenBurns;
    }
      break;

    }

  argc -= optind;
  argv += optind;

  if ((argc < 1)) {
    printHelpScreen(programName);
    return (-1);
  }

  /* create configuration */
  TheoraStreamParameter config;

  config.pictureX = width;
  config.pictureY = height;
  config.videoBitrate = datarate;
  config.aspectRatioDenom = 1;
  config.aspectRatioNum = 1;
  config.framerateNum = framesPerSecond;
  config.framerateDenom = 1;
  config.keyframeShift = 6;

  /* create stream configuration */
  TheoraEncoder theoraEncoder(0);
  StreamConfig streamConf;

  /* configure the theora encoder and get a stream config back
   * which configures the stream multiplexer */
  try {
  theoraEncoder.configureEncoder(config, streamConf);
  } catch(const char* data) {
    std::cerr << data;
    exit(-1);
  }

  // encoder might want another frame size:
  width = config.frameX;
  height = config.frameY;

  std::vector<StreamConfig> configList;
  configList.push_back(streamConf);

  /* create a repository, where the data should be placed */
  FileRepository repository(outputFile, MediaUnit::write);

  /* create a stream multiplexer */
  StreamMux streamCreate(&repository);

  /* configure the stream multiplexer */
  streamCreate.configureStreams(configList);

  /* extract the RGB picture plane */
  RGBPlane pictureRGB;
  RGBPlane pic2;

  for (int32 i(0); i<argc; ++i) {
    try {

      std::cout << "\ncreating video stream for picture <"<<argv[i]<<">\n";

      bool biggest = (!reframe);
      if (pictureRGB.load(argv[i], (uint32)(width*resample), (uint32)(height*resample), biggest) == false) {
        continue;

      }
      if (reframe)
    	  pictureRGB = pictureRGB.reframe((uint32)(width*resample), (uint32)(height*resample));

      switch(showType) {
        case seqType_KenBurns:
        KenBurnsSequence(streamCreate, theoraEncoder, pictureRGB, width, height, length, 1, framesPerSecond);
        break;

        case seqType_Crossfade:
        crossfadeSequence(streamCreate, theoraEncoder, pictureRGB, width, height, length, 1, framesPerSecond);
        break;

        case seqType_Plain:
        frameShowSequence(streamCreate, theoraEncoder, pictureRGB, width, height, length, 1, framesPerSecond);

      }

    }
    catch (const char* errorString) {
      std::cout << errorString << std::endl;
      exit(-1);
    }
  }

  pictureRGB.cleanup();

  streamCreate.setEndOfStream();

  std::cout << std::endl;
#ifdef OSX_MALLOC_DEBUG
  std::cout << "Done!\n";
  while (1==1) { }
#endif

  return(0);
}
